@@ -41,7 +41,6 @@ proc FTP_MakeSureLinkIsUp { ftpI } {
 }
 
 proc FTP_TrimDir { dir } {
-  set dir [string trim $dir]
   while { [string range $dir 0 1] == "//" } {
     set dir [string range $dir 1 end]
   }
@@ -49,8 +48,8 @@ proc FTP_TrimDir { dir } {
   if { $dir == "" } { set dir / }
   if { [string index $dir 0] == "/" } {
     while { 1 } {
-      set i [string last .. $dir]
-      if { $i != -1 } {
+      set len [string length $dir]
+      if { [string range $dir [expr $len - 3] end] == "/.." } {
         set dir [file dirname [file dirname $dir]]
       } else {
         break
@@ -75,33 +74,37 @@ proc FTP_CD { ftpI new_wd } {
 proc FTP_Rename { ftpI oldname newname } {
   global ftp
   FTP_MakeSureLinkIsUp $ftpI
-  set ftp($ftpI,rename,oldname) [string trim $oldname]
-  set ftp($ftpI,rename,newname) [string trim $newname]
+  set ftp($ftpI,rename,oldname) $oldname
+  set ftp($ftpI,rename,newname) $newname
   set ftp($ftpI,state) rename
+  FTP_InvalidateCache
   FTP_CtlAutomata $ftpI
 }
 
 proc FTP_Delete { ftpI filename } {
   global ftp
   FTP_MakeSureLinkIsUp $ftpI
-  set ftp($ftpI,delete,filename) [string trim $filename]
+  set ftp($ftpI,delete,filename) $filename
   set ftp($ftpI,state) delete
+  FTP_InvalidateCache
   FTP_CtlAutomata $ftpI
 }
 
 proc FTP_MkDir { ftpI dir } {
   global ftp
   FTP_MakeSureLinkIsUp $ftpI
-  set ftp($ftpI,mkdir,dir) [string trim $dir]
+  set ftp($ftpI,mkdir,dir) $dir
   set ftp($ftpI,state) mkdir
+  FTP_InvalidateCache
   FTP_CtlAutomata $ftpI
 }
 
 proc FTP_RmDir { ftpI dir } {
   global ftp
   FTP_MakeSureLinkIsUp $ftpI
-  set ftp($ftpI,rmdir,dir) [string trim $dir]
+  set ftp($ftpI,rmdir,dir) $dir
   set ftp($ftpI,state) rmdir
+  FTP_InvalidateCache
   FTP_CtlAutomata $ftpI
 }
 
@@ -136,9 +139,18 @@ proc FTP_CloseSession { ftpI } {
 proc FTP_List { ftpI showall } {
   global ftp
   FTP_MakeSureLinkIsUp $ftpI
+  LogStatusOnly "Reading ftp directory $ftp($ftpI,realhost)$ftp($ftpI,pwd)"
+  set cache_result [FTP_ReadCache $ftp($ftpI,realhost)$ftp($ftpI,pwd)]
+  if {$cache_result != ""} {
+    LogStatusOnly "Reading ftp directory $ftp($ftpI,realhost)$ftp($ftpI,pwd) -- done (found in cache)"
+    return $cache_result
+  }
   set ftp($ftpI,state) listing
   set ftp($ftpI,fileshow,all) $showall
-  return [FTP_CtlAutomata $ftpI]
+  set result [FTP_CtlAutomata $ftpI]
+  FTP_WriteCache $ftp($ftpI,realhost)$ftp($ftpI,pwd) $result
+  LogStatusOnly "Reading ftp directory $ftp($ftpI,realhost)$ftp($ftpI,pwd) -- done"
+  return $result
 }
 
 proc FTP_DoSearch { ftpI filename } {
@@ -190,9 +202,15 @@ proc FTP_PutFile { ftpI localFileName remoteFileName expectedSize } {
   set ftp($ftpI,remote_fname) $remoteFileName
   set ftp($ftpI,local_fname) $localFileName
   set ftp($ftpI,expected_size) $expectedSize
+  FTP_InvalidateCache
   return [FTP_CtlAutomata $ftpI]
 }
 
+proc FTP_InvalidateCache {} {
+  global ftp
+  set ftp(cache) ""
+}
+
 
 # ----------- Helper functions
 
@@ -335,6 +353,7 @@ proc FTP_CtlAutomata { ftpI } {
         } else {
           set ret 0
         }
+        set ftp($ftpI,pwd) ""
         set ftp($ftpI,state) ready
       }
       getfile {
@@ -584,7 +603,7 @@ proc FTP_ReadControl { ftpI var_ctl_mess var_ctl_code var_ctl_code0 var_ctl_code
   set ctl_code1 [string index $code 1]
   set ctl_code2 [string index $code 2]
   if { $ftp($ftpI,debug) } {
-    puts -nonewline "$ctl_mess"
+    puts "$ctl_mess"
   }
 }
 
@@ -619,7 +638,7 @@ proc FTP_ShutDown { ftpI } {
 proc FTP_WriteControl { ftpI output } {
   global ftp
   set ftp($ftpI,ctl_cmd) "$output"
-  set r [catch {FTP_WriteText $ftp($ftpI,ctl_fd) "$output\n"} out]
+  set r [catch {FTP_WriteText $ftp($ftpI,ctl_fd) "$output\r\n"} out]
   if {$r} {FTP_Error $ftpI $out}
   if { $ftp($ftpI,debug) } {
     puts ">> $output"
@@ -638,7 +657,7 @@ proc FTP_SetupDatareceiver { ftpI } {
   }
   set r [catch {FTP_CreateServerSocket $ftp($ftpI,local_ip)} t]
   if {$r} {FTP_Error $ftpI $t}
-  regexp {([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+),([0-9]+) ([0-9]+)} $t match a1 a2 a3 a4 p ftp($ftpI,server_fd)
+  regexp {([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+),([0-9]+) (.+)} $t match a1 a2 a3 a4 p ftp($ftpI,server_fd)
   return "PORT $a1,$a2,$a3,$a4,[FTP_ConvPortToNums $p]"
 }
 
@@ -683,12 +702,18 @@ proc FTP_TransferFile { ftpI mode } {
     set to_fd $fd1
   }
 
+  #fconfigure $from_fd -translation binary
+  #fconfigure $to_fd -translation binary
+
   set size 0
   if {$ftp($ftpI,resume)} {
     set size $ftp($ftpI,resume,pos)
   }
   set start_time [clock seconds]
-  set tl { {0 -1} {0 -1} {0 -1} {0 -1} {0 -1} {0 -1} {0 -1} {0 -1} {0 -1} {0 -1} {0 -1} {0 -1} }
+  set tl {}
+  for {set i 0} {$i < 30} {incr i} {
+    lappend tl {0 -1}
+  }
 
   set t_one [ClockMilliSeconds]
   while { 1 } {
@@ -705,22 +730,30 @@ proc FTP_TransferFile { ftpI mode } {
     if { $timesum <= 0.0 } { set timesum 1 }
     set speed [format "%.2f" [expr ($bytesum / ($timesum / 1000.0)) / 1024.0]]
     set speed_Bps [expr ($bytesum / ($timesum / 1000.0))]
-    if {$speed_Bps <= 0} {
-      set eta "?"
-    } else {
+    set eta "?"
+    set eta_abs "?"
+    if {$speed_Bps > 0} {
       set tmp [format "%.0f" [expr ($ftp($ftpI,expected_size) - $size) / $speed_Bps]]
       if { $tmp >= 0 } { 
         set eta [format "%02d:%02d" [expr $tmp / 60] [expr $tmp % 60]] 
-      } else {
-        set eta "?"
+        if { $config(dateformat) == "yymmdd" } {
+          set tmp_date "%y%m%d "
+        } else {
+          set tmp_date "%y%m%d "
+        } 
+        set tmp_s [clock seconds]
+        if { [clock format [expr $tmp_s + $tmp] -format "%y%m%d"] == [clock format $tmp_s -format "%y%m%d"] } {
+          set tmp_date ""
+        }
+        set eta_abs [clock format [expr $tmp_s + $tmp] -format "$tmp_date%R"]
       }
     }
     if {$ftp($ftpI,expected_size) > 0} {
-      LogStatusOnly "Transfer [file tail $ftp($ftpI,remote_fname)] : $size / $ftp($ftpI,expected_size) bytes ($speed kB/s, ETA $eta)"
+      LogStatusOnly "Transfer [file tail $ftp($ftpI,remote_fname)] : $size / $ftp($ftpI,expected_size) bytes ($speed kB/s, ETA $eta $eta_abs)"
     } else {
       LogStatusOnly "Transfer [file tail $ftp($ftpI,remote_fname)] : $size bytes ($speed kB/s)"
     }
-    wm iconname . "$eta [file tail $ftp($ftpI,remote_fname)]"
+    wm iconname . "$eta $eta_abs [file tail $ftp($ftpI,remote_fname)]"
     update
     if { $glob(abortcmd) } {
       wm iconname . "$oldiconname"
@@ -753,7 +786,7 @@ proc FTP_TransferFile { ftpI mode } {
     }
     if {$chunk > [expr 2 * $oldchunk]} {set chunk [expr 2 * $oldchunk]}
     if {$chunk < [expr $oldchunk / 2]} {set chunk [expr $oldchunk / 2]}
-    if {$chunk > 300} {set chunk 300}
+    if {$chunk > 900} {set chunk 900}
     if {$chunk < 1} {set chunk 1}
   }
   FTP_Close $from_fd
@@ -784,3 +817,29 @@ proc FTP_TransferFile { ftpI mode } {
   return $size
 }
 
+proc FTP_ReadCache { key } {
+  global ftp
+  set i 0
+  foreach k $ftp(cache) {
+    if {[lindex $k 0] == "$key"} {
+      set item $k
+      set result [lindex $item 1]
+      set ftp(cache) [concat [lrange $ftp(cache) 0 [expr $i - 1]] [lrange $ftp(cache) [expr $i + 1] end]]
+      lappend ftp(cache) $item
+      return [lindex $item 1]
+    }
+    incr i
+  }
+  return ""
+}
+
+proc FTP_WriteCache { key data } {
+  global ftp config
+  set item [list $key $data]
+  lappend ftp(cache) $item
+  set length [llength $ftp(cache)]
+  if {$length > $config(ftp,cache,maxentries)} {
+    set ftp(cache) [lrange $ftp(cache) 1 end]
+  }
+}
+
